import json
import time

from openai import OpenAI
from tenacity import retry, wait_random_exponential, stop_after_attempt
import sys
from TPM import main
from datetime import datetime, timedelta
GPT_MODEL = "gpt-4-turbo"
client = OpenAI(api_key="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")

def count_date(period):
    today = datetime.today() - timedelta(days=1)
    half_year_ago = today - timedelta(days=period)
    half_year_ago_formatted = half_year_ago.strftime("%Y-%m-%d")
    return today.strftime("%Y-%m-%d"), half_year_ago_formatted
@retry(wait=wait_random_exponential(multiplier=1, max=40), stop=stop_after_attempt(3))
def chat_completion_request(messages, tools=None, tool_choice=None, model=GPT_MODEL):
    try:
        response = client.chat.completions.create(
            model=model,
            messages=messages,
            tools=tools,
            tool_choice=tool_choice,
        )
        return response
    except Exception as e:
        print("Unable to generate ChatCompletion response")
        print(f"Exception: {e}")
        sys.exit(1)  # Add this line

def get_tpm_backtest(symbols, sliding_window=126, frequency=126, function="max_sharpe",period="2022-01-01 to 2022-12-31"):
    if period.isalnum() == False:
        period = period.split(" to ")
        start_date = period[0]
        end_date = period[1]
    else:
        end_date, start_date  = count_date(int(period))
    print(symbols, sliding_window, frequency, function, period , start_date, end_date)
    result = main(symbols, role=function, start_date=start_date, end_date=end_date, lookback=int(sliding_window), backtest=int(frequency))
    if result == False:
        return "投資組合無法建立，資料長度與所選參數不符。"
    return f"Backtest result for {symbols} Annualized return: {result['annual_ret']}, Sharpe ratio: {result['annual_sr']}, Annualized volatility: {result['vol']}, Maximum drawdown: {result['mdd']}, for the period {start_date} to {end_date}"
def backest_main(query):
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_backtest",
                "description": "Get the portfolio backtesting result by combined with a list of symbol, sliding window, optimize frequency and  optimize function",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "symbol": {
                            "type": "array",
                            "description": "An array of multiple portfolio symbol to be backtested if the symbol is Taiwan Stock exchage the code ex: TSMC to 2330.TW ,GOOGLE to GOOG , APPLE to AAPL , if there is multiple symbol return a python list format",
                            "items": {
                                "type": "string",
                                "description": "The symbol of the stock",
                            },
                        },
                        "sliding window": {
                            "type": "string",
                            "enum": ["21", "63", "126", "252"],
                            "description": "The sliding window size to be backtested in one month:21, three months:63 , six months:126, one year:252",
                        },
                        "frequency": {
                            "type": "string",
                            "enum": ["21", "63", "126", "252"],
                            "description": "The optimize frequency to be backtested in monthly:21, quarterly:63, semi-annually:126, annually:252",
                        },
                        "function": {
                            "type": "string",
                            "enum": ["max_sharpe", "max_sortino", "min_volatility", "quadratic_utility"],
                            "description": "The optimize function to be backtested in sharpe ratio, sortino ratio,volatility, utility function",
                        },
                        "period": {
                            "type": "string",
                            "description": "The period of the backtest to be calculated If the user specifies a time period for backtesting, return it in the format ‘YYYY-MM-DD to YYYY-MM-DD’. If the user specifies the past year, return ‘360’. If the user specifies the past six months, return ‘180",
                        },
                    },
                    "required": ["symbol", "sliding window", "frequency", "function","period"]
                },
            }
        }
    ]

    messages = []
    sec_message = []
    messages.append({"role": "system", "content": "You are a software developer who is writing a function to get the portfolio backtesting result by using different multiple symbol, sliding window, optimize frequency and  optimize function , "
                                                  "only the symbol is required, the sliding window, frequency and function are optional. The sliding window size can be default by six months. The optimize frequency can be default by semi-annually. The optimize function can be default by sharpe ratio. "})
    messages.append({"role": "user", "content": query})

    chat_response = chat_completion_request(messages,tools=tools)

    assistant_messages = chat_response.choices[0].message.tool_calls
    messages.append(assistant_messages)
    available_functions = {
        "get_backtest": get_tpm_backtest,
    }  # only one function in this example, but you can have multiple
    for tool_call in assistant_messages:
        function_name = tool_call.function.name
        function_to_call = available_functions[function_name]
        function_args = json.loads(tool_call.function.arguments)
        function_response = function_to_call(
            symbols=function_args.get("symbol"),
            sliding_window=function_args.get("sliding window"),
            frequency=function_args.get("frequency"),
            function=function_args.get("function"),
            period=function_args.get("period")
        )
        sec_message.append({"role": "system","content": "You are a professional financial analyst. The user will provide you with some results from their backtesting of an investment portfolio using the Efficient Frontier calculation. These results include annualized return, annualized Sharpe ratio, annualized volatility, maximum drawdown. Please provide professional advice in 繁體中文 based on these reports. "})
        sec_message.append({"role": "user","content": function_response})
        result_messages =  chat_completion_request(sec_message)
    return result_messages.choices[0].message.content

query = "我想要使用GOOGLE和台積電來進行最大夏普比率2022/7/1至2024/6/01的回測"
article = backest_main(query)
print(article)